import ipaddress
import os
import re
import socket
import sys
from datetime import datetime
from importlib import metadata
from time import sleep

import pymysql
import requests
from dolphie.Database import Database
from dolphie.Functions import detect_encoding, format_bytes, format_number
from dolphie.KBHit import KBHit
from dolphie.Queries import Queries
from packaging.version import parse as parse_version
from rich import box
from rich.align import Align
from rich.console import Console
from rich.layout import Layout
from rich.style import Style
from rich.syntax import Syntax
from rich.table import Table
from rich.text import Text

try:
    __package_name__ = metadata.metadata(__package__ or __name__)["Name"]
    __version__ = metadata.version(__package__ or __name__)
except Exception:
    __package_name__ = "N/A"
    __version__ = "N/A"


class Dolphie:
    def __init__(self):
        self.console = Console()
        self.kb = KBHit()
        self.rich_live = None

        # Config options
        self.user: str = None
        self.password: str = None
        self.host: str = None
        self.port: int = 3306
        self.socket: str = None
        self.ssl: dict = {}
        self.config_file: str = None
        self.host_cache_file: str = None
        self.debug: bool = False
        self.refresh_interval: int = 1
        self.refresh_interval_innodb_status: int = 1
        self.dashboard: bool = True
        self.use_processlist: bool = False
        self.show_idle_queries: bool = False
        self.show_trxs_only: bool = False
        self.show_additional_query_columns: bool = False
        self.show_last_executed_query: bool = False
        self.sort_by_time_descending: bool = True
        self.heartbeat_table: str = None
        self.user_filter: str = None
        self.db_filter: str = None
        self.host_filter: str = None
        self.time_filter: str = 0
        self.query_filter: str = None

        # Loop variables
        self.dolphie_start_time: datetime = datetime.now()
        self.previous_main_loop_time: datetime = datetime.now()
        self.previous_innodb_status_loop_time: datetime = datetime.now()
        self.loop_duration_seconds: int = 0
        self.processlist_threads: dict = {}
        self.pause_refresh: bool = False
        self.first_loop: bool = True
        self.saved_status: bool = None
        self.previous_binlog_position: int = 0
        self.previous_replica_sbm: int = 0
        self.host_cache: dict = {}
        self.variables: dict = {}
        self.statuses: dict = {}
        self.primary_status: dict = {}
        self.replica_status: dict = {}
        self.innodb_status: dict = {}
        self.replica_connections: dict = {}

        # Set on database connection
        self.connection_id: int = None
        self.use_performance_schema: bool = False
        self.performance_schema_enabled: bool = False
        self.innodb_locks_sql: bool = False
        self.host_is_rds: bool = False
        self.server_uuid: str = None
        self.full_version: str = None
        self.host_distro: str = None

        self.app_version = __version__
        self.base_title = f"[b steel_blue1]dolphie :dolphin: {self.app_version}[/b steel_blue1]"
        self.title = f"{self.base_title}[grey62] press ? for help"

    def check_for_update(self):
        # Query PyPI API to get the latest version
        url = f"https://pypi.org/pypi/{__package_name__}/json"
        response = requests.get(url)

        if response.status_code == 200:
            data = response.json()

            # Extract the latest version from the response
            latest_version = data["info"]["version"]

            # Compare the current version with the latest version
            if parse_version(latest_version) > parse_version(__version__):
                self.console.print(
                    (
                        "[bright_green]New version available!\n\n[grey93]Current version:"
                        f" [steel_blue1]{__version__}\n[grey93]Latest version:"
                        f" [steel_blue1]{latest_version}\n\n[grey93]Please update to the latest version at your"
                        " convenience\n\n[steel_blue1]Press any key to continue"
                    ),
                    highlight=False,
                )

                key = self.kb.key_press_blocking()
                if key:
                    pass
        else:
            self.console.print(
                f"[bright_red]Failed to retrieve package information from PyPI![/bright_red] URL: {url} - Code:"
                f" {response.status_code}"
            )

    def create_rich_layout(self) -> Layout:
        layout = Layout(name="root")

        layout.split_column(
            Layout(name="header", size=1),
            Layout(name="dashboard", size=13),
            Layout(name="innodb_io", visible=False, size=10),
            Layout(name="replicas", visible=False),
            Layout(name="innodb_locks", visible=False),
            Layout(name="processlist"),
            Layout(name="footer", size=1, visible=False),
        )

        layout["header"].update(Align.right(Text.from_markup(self.base_title)))
        layout["dashboard"].update("")
        layout["processlist"].update("")
        layout["footer"].update("")

        self.layout = layout

    def update_footer(self, output):
        self.layout["footer"].visible = True

        self.layout["footer"].update(output)
        self.rich_live.update(self.layout, refresh=True)
        sleep(1.5)
        self.layout["footer"].update("")

        self.layout["footer"].visible = False

    def db_connect(self):
        self.db = Database(self.host, self.user, self.password, self.socket, self.port, self.ssl)

        query = "SELECT CONNECTION_ID() AS connection_id"
        self.connection_id = self.db.fetchone(query, "connection_id")

        query = "SELECT @@performance_schema"
        performance_schema = self.db.fetchone(query, "@@performance_schema")
        if performance_schema == 1:
            self.performance_schema_enabled = True

            if not self.use_processlist:
                self.use_performance_schema = True

        query = "SELECT @@version_comment"
        version_comment = self.db.fetchone(query, "@@version_comment").lower()

        query = "SELECT @@basedir"
        basedir = self.db.fetchone(query, "@@basedir")

        self.db.cursor.execute("SHOW GLOBAL VARIABLES LIKE 'aurora_version'")
        data = self.db.cursor.fetchone()
        aurora_version = None
        if data:
            aurora_version = data["Value"].decode()

        query = "SELECT @@version"
        version = self.db.fetchone(query, "@@version").lower()
        version_split = version.split(".")

        self.full_version = "%s.%s.%s" % (
            version_split[0],
            version_split[1],
            version_split[2].split("-")[0],
        )

        major_minor_version = "%s.%s" % (version_split[0], version_split[1])
        major_version = version_split[0]

        # Get proper host version and fork
        if "percona xtradb cluster" in version_comment:
            self.host_distro = "Percona XtraDB Cluster"
        elif "percona server" in version_comment:
            self.host_distro = "Percona Server"
        elif "mariadb cluster" in version_comment:
            self.host_distro = "MariaDB Cluster"
        elif "mariadb" in version_comment or "mariadb" in version:
            self.host_distro = "MariaDB"
        elif aurora_version:
            self.host_distro = "Amazon Aurora"
            self.host_is_rds = True
        elif "rdsdb" in basedir:
            self.host_distro = "Amazon RDS"
            self.host_is_rds = True
        else:
            self.host_distro = "MySQL"

        # Determine if InnoDB locks panel is available for a version and which query to use
        # Major version 10 is for MariaDB
        self.innodb_locks_sql = None
        innodb_locks_queries = {
            "5.6": Queries["locks_query-5"],
            "5.7": Queries["locks_query-5"],
            "8": Queries["locks_query-8"] if self.use_performance_schema else None,
            "10": Queries["locks_query-5"],
        }
        if major_minor_version in innodb_locks_queries or major_version in innodb_locks_queries:
            self.innodb_locks_sql = innodb_locks_queries.get(major_minor_version) or innodb_locks_queries.get(
                major_version
            )

        query = "SELECT @@server_uuid"
        if "MariaDB" in self.host_distro:
            query = "SELECT @@server_id AS @@server_uuid"
        self.server_uuid = self.db.fetchone(query, "@@server_uuid")

    def fetch_data(self, command):
        command_data = {}

        if command == "status" or command == "variables":
            self.db.cursor.execute(Queries[command])
            data = self.db.cursor.fetchall()

            for row in data:
                variable = row["Variable_name"].decode()
                value = row["Value"]

                try:
                    converted_value = row["Value"].decode()

                    if converted_value.isnumeric():
                        converted_value = int(converted_value)
                except (UnicodeDecodeError, AttributeError):
                    converted_value = value

                command_data[variable] = converted_value

        elif command == "innodb_status":
            self.db.cursor.execute(Queries[command])
            data = self.db.cursor.fetchone()

            command_data["status"] = data["Status"].decode(detect_encoding(data["Status"]))

        else:
            self.db.cursor.execute(Queries[command])
            data = self.db.cursor.fetchall()

            for row in data:
                for column, value in row.items():
                    try:
                        converted_value = value.decode()

                        if converted_value.isnumeric():
                            converted_value = int(converted_value)
                    except (UnicodeDecodeError, AttributeError):
                        converted_value = value

                    command_data[column] = converted_value

        return command_data

    def block_refresh_for_key_command(self, refresh=False):
        if refresh:
            self.layout["footer"].visible = True
            self.layout["footer"].update(Align.center("[b]Paused![/b] Press any key to resume", style="steel_blue1"))
            self.rich_live.update(self.layout, refresh=True)
        else:
            self.console.print(Align.center("\n[b]Paused![/b] Press any key to resume", style="steel_blue1"))

        key = self.kb.key_press_blocking()
        if key:
            self.layout["footer"].update("")
            self.layout["footer"].visible = False
            return

    def capture_key(self, key):
        self.pause_refresh = True
        valid_key = True

        self.kb.set_normal_term()

        if key == "1":
            if self.use_performance_schema:
                self.use_performance_schema = False
                self.update_footer("[steel_blue1]Switched to [grey93]Processlist")
            else:
                if self.performance_schema_enabled:
                    self.use_performance_schema = True
                    self.update_footer("[steel_blue1]Switched to [grey93]Performance Schema")
                else:
                    self.update_footer("[red]You can't switch to Performance Schema because it isn't enabled")

        elif key == "2":
            os.system("clear")
            self.console.print(Align.center(self.title))

            self.console.print(self.innodb_status["status"])

            self.block_refresh_for_key_command()

        elif key == "a":
            if self.show_additional_query_columns:
                self.show_additional_query_columns = False
            else:
                self.show_additional_query_columns = True

        elif key == "c":
            self.user_filter = ""
            self.db_filter = ""
            self.host_filter = ""
            self.time_filter = ""
            self.query_filter = ""

            self.update_footer("[steel_blue1]Cleared all filters!")

        elif key == "d":
            if self.layout["dashboard"].visible:
                self.layout["dashboard"].visible = False
            else:
                self.layout["dashboard"].visible = True

        elif key == "D":
            self.db_filter = self.console.input("[steel_blue1]Database to filter by[/steel_blue1]: ")

        elif key == "e":
            thread_id = self.console.input("[steel_blue1]Thread ID to explain[/steel_blue1]: ")

            if thread_id:
                if thread_id in self.processlist_threads:
                    os.system("clear")
                    self.console.print(Align.right(self.title))

                    row_style = Style(color="grey93")
                    table = Table(box=box.ROUNDED, show_header=False, style="grey70")
                    table.add_column("")
                    table.add_column("")

                    table.add_row("[gray78]Thread ID", str(thread_id), style=row_style)
                    table.add_row(
                        "[gray78]User",
                        self.processlist_threads[thread_id]["user"],
                        style=row_style,
                    )
                    table.add_row(
                        "[gray78]Host",
                        self.processlist_threads[thread_id]["host"],
                        style=row_style,
                    )
                    table.add_row(
                        "[gray78]Database",
                        self.processlist_threads[thread_id]["db"],
                        style=row_style,
                    )
                    table.add_row(
                        "[gray78]Command",
                        self.processlist_threads[thread_id]["command"],
                        style=row_style,
                    )
                    table.add_row(
                        "[gray78]State",
                        self.processlist_threads[thread_id]["state"],
                        style=row_style,
                    )
                    table.add_row(
                        "[gray78]Time",
                        self.processlist_threads[thread_id]["hhmmss_time"],
                        style=row_style,
                    )
                    table.add_row(
                        "[gray78]Rows Locked",
                        self.processlist_threads[thread_id]["trx_rows_locked"],
                        style=row_style,
                    )
                    table.add_row(
                        "[gray78]Rows Modified",
                        self.processlist_threads[thread_id]["trx_rows_modified"],
                        style=row_style,
                    )
                    if "innodb_thread_concurrency" in self.variables and self.variables["innodb_thread_concurrency"]:
                        table.add_row(
                            "[gray78]Tickets",
                            self.processlist_threads[thread_id]["trx_concurrency_tickets"],
                            style=row_style,
                        )
                    table.add_row("", "")
                    table.add_row(
                        "[gray78]TRX State",
                        self.processlist_threads[thread_id]["trx_state"],
                        style=row_style,
                    )
                    table.add_row(
                        "[gray78]TRX Operation",
                        self.processlist_threads[thread_id]["trx_operation_state"],
                        style=row_style,
                    )

                    self.console.print(Align.center(table))

                    if self.processlist_threads[thread_id]["query"]:
                        self.console.print("")
                        self.console.print(
                            Align.center(
                                Syntax(
                                    self.processlist_threads[thread_id]["query"],
                                    "sql",
                                    line_numbers=False,
                                    word_wrap=True,
                                    theme="vim",
                                )
                            )
                        )

                    if self.processlist_threads[thread_id]["query"] and self.processlist_threads[thread_id]["db"]:
                        explain_failure = None
                        explain_data = None

                        self.console.print("")

                        try:
                            self.db.cursor.execute("USE %s" % self.processlist_threads[thread_id]["db"])
                            self.db.cursor.execute("EXPLAIN %s" % self.processlist_threads[thread_id]["query"])

                            explain_data = self.db.cursor.fetchall()
                        except pymysql.Error as e:
                            explain_failure = "[bright_red]EXPLAIN ERROR:[/bright_red] [red]%s" % e.args[1]

                        if explain_data:
                            explain_table = Table(box=box.ROUNDED, style="grey70")

                            columns = []
                            for row in explain_data:
                                values = []
                                for column, value in row.items():
                                    # Exclude possbile_keys field since it takes up too much space
                                    if column == "possible_keys":
                                        continue

                                    # Don't duplicate columns
                                    if column not in columns:
                                        explain_table.add_column(column)
                                        columns.append(column)

                                    if column == "key" and value is None:
                                        value = "[b bright_red]NONE[/b bright_red]"

                                    if column == "rows":
                                        value = format_number(value)

                                    try:
                                        values.append(value.decode())
                                    except (UnicodeDecodeError, AttributeError):
                                        values.append(str(value))

                                explain_table.add_row(*values, style="grey93")

                            self.console.print(Align.center(explain_table))
                        else:
                            self.console.print(Align.center(explain_failure))

                    self.block_refresh_for_key_command()
                else:
                    self.update_footer("[bright_red]Thread ID '[grey93]%s[bright_red]' does not exist!" % thread_id)

        elif key == "H":
            self.host_filter = self.console.input("[steel_blue1]Hostname/IP to filter by[/steel_blue1]: ")

            # Since our filtering is done by the processlist query, the value needs to be what's in host cache
            for ip, addr in self.host_cache.items():
                if self.host_filter == addr:
                    self.host_filter = ip

        elif key == "i":
            if self.layout["innodb_io"].visible:
                self.layout["innodb_io"].visible = False
            else:
                self.layout["innodb_io"].visible = True

        elif key == "I":
            if self.show_idle_queries:
                self.show_idle_queries = False
                self.sort_by_time_descending = True
            else:
                self.show_idle_queries = True
                self.sort_by_time_descending = False

        elif key == "k":
            thread_id = self.console.input("[steel_blue1]Thread ID to kill[/steel_blue1]: ")

            if thread_id:
                if thread_id in self.processlist_threads:
                    try:
                        if self.host_is_rds:
                            self.db.cursor.execute("CALL mysql.rds_kill(%s)" % thread_id)
                        else:
                            self.db.cursor.execute("KILL %s" % thread_id)
                    except pymysql.OperationalError:
                        self.update_footer("[bright_red]Thread ID '[grey93]%s[bright_red]' does not exist!" % thread_id)
                else:
                    self.update_footer("[bright_red]Thread ID '[grey93]%s[bright_red]' does not exist!" % thread_id)

        elif key == "K":
            include_sleep = self.console.input("[steel_blue1]Include queries in sleep state? (y/n)[/steel_blue1]: ")

            if include_sleep != "y" and include_sleep != "n":
                self.update_footer("[bright_red]Invalid option!")
            else:
                kill_type = self.console.input(
                    "[steel_blue1]Kill by username/hostname/time range (u/h/t)[/steel_blue1]: "
                )
                threads_killed = 0

                commands_without_sleep = ["Query", "Execute"]
                commands_with_sleep = ["Query", "Execute", "Sleep"]

                if kill_type == "u":
                    user = self.console.input("[steel_blue1]User[/steel_blue1]: ")

                    for thread_id, thread in self.processlist_threads.items():
                        try:
                            if thread["user"] == user:
                                if include_sleep == "y":
                                    if thread["command"] in commands_with_sleep:
                                        if self.host_is_rds:
                                            self.db.cursor.execute("CALL mysql.rds_kill(%s)" % thread_id)
                                        else:
                                            self.db.cursor.execute("KILL %s" % thread_id)

                                        threads_killed += 1
                                else:
                                    if thread["command"] in commands_without_sleep:
                                        if self.host_is_rds:
                                            self.db.cursor.execute("CALL mysql.rds_kill(%s)" % thread_id)
                                        else:
                                            self.db.cursor.execute("KILL %s" % thread_id)

                                        threads_killed += 1
                        except pymysql.OperationalError:
                            continue

                elif kill_type == "h":
                    host = self.console.input("[steel_blue1]Host/IP[/steel_blue1]: ")

                    for thread_id, thread in self.processlist_threads.items():
                        try:
                            if thread["host"] == host:
                                if include_sleep == "y":
                                    if thread["command"] in commands_with_sleep:
                                        if self.host_is_rds:
                                            self.db.cursor.execute("CALL mysql.rds_kill(%s)" % thread_id)
                                        else:
                                            self.db.cursor.execute("KILL %s" % thread_id)

                                        threads_killed += 1
                                else:
                                    if thread["command"] in commands_without_sleep:
                                        if self.host_is_rds:
                                            self.db.cursor.execute("CALL mysql.rds_kill(%s)" % thread_id)
                                        else:
                                            self.db.cursor.execute("KILL %s" % thread_id)

                                        threads_killed += 1
                        except pymysql.OperationalError:
                            continue

                elif kill_type == "t":
                    time = self.console.input("[steel_blue1]Time range (ex. 10-20)[/steel_blue1]: ")

                    if re.search(r"(\d+-\d+)", time):
                        time_range = time.split("-")
                        lower_limit = int(time_range[0])
                        upper_limit = int(time_range[1])

                        if lower_limit > upper_limit:
                            self.update_footer(
                                "[bright_red]Invalid time range! Lower limit can't be higher than upper!"
                            )
                        else:
                            for thread_id, thread in self.processlist_threads.items():
                                try:
                                    if thread["time"] >= lower_limit and thread["time"] <= upper_limit:
                                        if include_sleep == "y":
                                            if thread["command"] in commands_with_sleep:
                                                if self.host_is_rds:
                                                    self.db.cursor.execute("CALL mysql.rds_kill(%s)" % thread_id)
                                                else:
                                                    self.db.cursor.execute("KILL %s" % thread_id)

                                                threads_killed += 1
                                        else:
                                            if thread["command"] in commands_without_sleep:
                                                if self.host_is_rds:
                                                    self.db.cursor.execute("CALL mysql.rds_kill(%s)" % thread_id)
                                                else:
                                                    self.db.cursor.execute("KILL %s" % thread_id)

                                                threads_killed += 1
                                except pymysql.OperationalError:
                                    continue
                    else:
                        self.update_footer("[bright_red]Invalid time range!")
                else:
                    self.update_footer("[bright_red]Invalid option!")

                if threads_killed:
                    self.update_footer("[grey93]Killed [steel_blue1]%s [grey93]threads!" % threads_killed)
                else:
                    self.update_footer("[bright_red]No threads were killed!")

        elif key == "l":
            if self.innodb_locks_sql:
                if self.layout["innodb_locks"].visible:
                    self.layout["innodb_locks"].visible = False
                else:
                    self.layout["innodb_locks"].visible = True
            else:
                self.update_footer("[red]Lock panel isn't supported for this host's version!")

        elif key == "L":
            os.system("clear")
            self.console.print(Align.right(self.title))

            deadlock = ""
            output = re.search(
                r"------------------------\nLATEST\sDETECTED\sDEADLOCK\n------------------------"
                "\n(.*?)------------\nTRANSACTIONS",
                self.innodb_status["status"],
                flags=re.S,
            )
            if output:
                deadlock = output.group(1)

                deadlock = deadlock.replace("***", "[yellow]*****[/yellow]")
                self.console.print("[steel_blue1]Latest deadlock detected:")
                self.console.print(deadlock, highlight=False)
            else:
                self.console.print("No deadlock detected!", justify="center")

            self.block_refresh_for_key_command()

        elif key == "p":
            if self.layout["processlist"].visible:
                self.layout["processlist"].visible = False
            else:
                self.layout["processlist"].visible = True

        elif key == "P":
            self.block_refresh_for_key_command(
                refresh=True,
            )

        elif key == "Q":
            self.query_filter = self.console.input("[steel_blue1]Query text to filter by[/steel_blue1]: ")

        elif key == "q":
            sys.exit()

        elif key == "r":
            if self.layout["replicas"].visible:
                self.layout["replicas"].visible = False

                # Cleanup connections
                for connection in self.replica_connections.values():
                    connection["connection"].close()

                self.replica_connections = {}
            else:
                self.layout["replicas"].visible = True

        elif key == "s":
            if self.sort_by_time_descending:
                self.sort_by_time_descending = False
            else:
                self.sort_by_time_descending = True

        elif key == "S":
            if self.show_last_executed_query:
                self.show_last_executed_query = False
            else:
                self.show_last_executed_query = True

        elif key == "t":
            if self.show_trxs_only:
                self.show_trxs_only = False
            else:
                self.show_trxs_only = True

        elif key == "T":
            time = self.console.input("[steel_blue1]Minimum time to display for queries[/steel_blue1]: ")

            if time.isnumeric():
                self.time_filter = int(time)
            else:
                self.update_footer("[bright_red]Time must be an integer!")

        elif key == "u":
            user_stat_data = self.create_user_stats_table()
            if user_stat_data:
                os.system("clear")
                self.console.print(Align.right(self.title))
                self.console.print(Align.center(user_stat_data))

                self.block_refresh_for_key_command()
            else:
                self.update_footer(
                    "[bright_red]This feature requires Userstat variable or Performance Schema to be enabled!"
                )

        elif key == "U":
            self.user_filter = self.console.input("[steel_blue1]User to filter by[/steel_blue1]: ")

        elif key == "y":
            os.system("clear")

            db_counter = 1
            row_counter = 1
            table_counter = 1
            tables = {}
            all_tables = []

            db_count = self.db.cursor.execute(Queries["databases"])
            databases = self.db.cursor.fetchall()

            # Determine how many tables to provide data
            if db_count <= 20:
                max_num_tables = 1
            else:
                max_num_tables = 3

            # Calculate how many databases per table
            row_per_count = round(db_count / max_num_tables)

            # Create dictionary of how many tables we want
            table_grid = Table.grid()
            while table_counter <= max_num_tables:
                tables[table_counter] = Table(box=box.ROUNDED, show_header=False, style="grey70")
                tables[table_counter].add_column("")

                table_counter += 1

            # Loop databases
            table_counter = 1
            for database in databases:
                tables[table_counter].add_row(database["Database"].decode(), style="grey93")

                if db_counter == row_per_count and row_counter != max_num_tables:
                    row_counter += 1
                    db_counter = 0
                    table_counter += 1

                db_counter += 1

            # Put all the variable data from dict into an array
            for table, table_data in tables.items():
                if table_data:
                    all_tables.append(table_data)

            table_grid.add_row(*all_tables)

            self.console.print(Align.right(self.title))
            self.console.print(Align.center(table_grid))
            self.console.print(Align.center("Total: [b steel_blue1]%s" % db_count))

            self.block_refresh_for_key_command()

        elif key == "v":
            input_variable = self.console.input(
                "[steel_blue1]Variable wildcard search ([grey78]leave blank for all[steel_blue1])[/steel_blue1]: "
            )

            os.system("clear")

            size_convert_variables = [
                "audit_log_buffer_size",
                "audit_log_rotate_on_size",
                "binlog_cache_size",
                "binlog_row_event_max_size",
                "binlog_stmt_cache_size",
                "bulk_insert_buffer_size",
                "clone_buffer_size",
                "delay_queue_size",
                "histogram_generation_max_mem_size",
                "innodb_buffer_pool_chunk_size",
                "innodb_buffer_pool_size",
                "innodb_change_buffer_max_size",
                "innodb_ft_cache_size",
                "innodb_ft_total_cache_size",
                "innodb_log_buffer_size",
                "innodb_log_file_size",
                "innodb_log_write_ahead_size",
                "innodb_max_bitmap_file_size",
                "innodb_max_undo_log_size",
                "innodb_online_alter_log_max_size",
                "innodb_page_size",
                "innodb_sort_buffer_size",
                "join_buffer_size",
                "key_buffer_size",
                "key_cache_block_size",
                "large_page_size",
                "max_binlog_cache_size",
                "max_binlog_size",
                "max_binlog_stmt_cache_size",
                "max_heap_table_size",
                "max_join_size",
                "max_relay_log_size",
                "myisam_data_pointer_size",
                "myisam_max_sort_file_size",
                "myisam_mmap_size",
                "myisam_sort_buffer_size",
                "optimizer_trace_max_mem_size",
                "parser_max_mem_size",
                "preload_buffer_size",
                "query_alloc_block_size",
                "query_prealloc_size",
                "range_alloc_block_size",
                "range_optimizer_max_mem_size",
                "read_buffer_size",
                "read_rnd_buffer_size",
                "rpl_read_size",
                "slave_pending_jobs_size_max",
                "sort_buffer_size",
                "tmp_table_size",
                "transaction_alloc_block_size",
                "transaction_prealloc_size",
                "max_allowed_packet",
                "innodb_redo_log_capacity",
                "replica_max_allowed_packet",
                "mysqlx_max_allowed_packet",
                "global_connection_memory_limit",
                "temptable_max_mmap",
                "temptable_max_ram",
            ]

            table_grid = Table.grid()
            table_counter = 1
            variable_counter = 1
            row_counter = 1
            variable_num = 1
            all_tables = []
            tables = {}
            display_variables = {}

            for variable, value in self.variables.items():
                if input_variable:
                    if input_variable not in variable:
                        continue

                # Convert size variables so they're readable
                if variable in size_convert_variables:
                    try:
                        display_variables[variable] = format_bytes(self.variables[variable])
                    except KeyError:
                        display_variables[variable] = "N/A"
                else:
                    display_variables[variable] = value

            max_num_tables = 1 if len(display_variables) <= 20 else 3

            # Create the number of tables we want
            while table_counter <= max_num_tables:
                tables[table_counter] = Table(box=box.ROUNDED, show_header=False, style="grey70")
                tables[table_counter].add_column("")
                tables[table_counter].add_column("")

                table_counter += 1

            # Calculate how many variables per table
            row_per_count = round(len(display_variables) / max_num_tables)

            # Loop variables
            for variable, value in display_variables.items():
                tables[variable_num].add_row("[gray78]%s" % variable, str(value), style="grey93")

                if variable_counter == row_per_count and row_counter != max_num_tables:
                    row_counter += 1
                    variable_counter = 0
                    variable_num += 1

                variable_counter += 1

            # Put all the variable data from dict into an array
            all_tables = [table_data for table_data in tables.values() if table_data]

            self.console.print(Align.right(self.title))

            # Add the data into a single tuple for add_row
            if display_variables:
                table_grid.add_row(*all_tables)
                self.console.print(Align.center(table_grid))
                self.block_refresh_for_key_command()
            else:
                if input_variable:
                    self.update_footer("[bright_red]No variable(s) found that match your search!")

        elif key == "x":
            refresh_interval = self.console.input("[steel_blue1]Refresh interval (in seconds)[/steel_blue1]: ")

            if refresh_interval.isnumeric():
                self.refresh_interval = int(refresh_interval)
            else:
                self.update_footer("[bright_red]Input must be an integer!")

        elif key == "X":
            refresh_interval_innodb_status = self.console.input(
                "[steel_blue1]Refresh interval for InnoDB Status data (in seconds)[/steel_blue1]: "
            )

            if refresh_interval_innodb_status.isnumeric():
                self.refresh_interval_innodb_status = int(refresh_interval_innodb_status)
            else:
                self.update_footer("[bright_red]Input must be an integer!")

        elif key == "z":
            os.system("clear")
            self.console.print(Align.right(self.title))

            if self.host_cache:
                table = Table(box=box.ROUNDED, style="grey70")
                table.add_column("IP")
                table.add_column("Hostname")

                for ip, addr in self.host_cache.items():
                    table.add_row(ip, addr)

                self.console.print(Align.center(table))
                self.console.print(Align.center("Total: [b steel_blue1]%s" % len(self.host_cache)))
            else:
                self.console.print(Align.center("\nThere are currently no IPs resolved to a hostname"))

            self.block_refresh_for_key_command()

        elif key == "?":
            os.system("clear")

            row_style = Style(color="grey93")

            filters = {
                "c": "Clear all filters",
                "D": "Filter by database",
                "H": "Filter by host/IP",
                "Q": "Filter by query text",
                "T": "Filter by minimum query time",
                "U": "Filter by user",
            }
            table_filters = Table(box=box.ROUNDED, style="grey70", title="Filters", title_style="bold steel_blue1")
            table_filters.add_column("Key", justify="center", style="b steel_blue1")
            table_filters.add_column("Description")
            for key, description in sorted(filters.items()):
                table_filters.add_row("[steel_blue1]%s" % key, description, style=row_style)

            panels = {
                "d": "Show/hide dashboard",
                "i": "Show/hide InnoDB information",
                "l": "Show/hide InnoDB query locks",
                "p": "Show/hide processlist",
                "r": "Show/hide replication",
            }
            table_panels = Table(box=box.ROUNDED, style="grey70", title="Panels", title_style="bold steel_blue1")
            table_panels.add_column("Key", justify="center", style="b steel_blue1")
            table_panels.add_column("Description")
            for key, description in sorted(panels.items()):
                table_panels.add_row("[steel_blue1]%s" % key, description, style=row_style)

            keys = {
                "1": "Switch between using Processlist/Performance Schema for listing queries",
                "2": "Print output from SHOW ENGINE INNODB STATUS",
                "a": "Show/hide additional processlist columns",
                "e": "Explain query of a thread and display thread information",
                "I": "Show/hide idle queries",
                "k": "Kill a query by thread ID",
                "K": "Kill a query by either user/host/time range",
                "L": "Show latest deadlock detected",
                "P": "Pause Dolphie",
                "q": "Quit Dolphie",
                "t": "Show/hide running transactions only",
                "s": "Sort query list by time in descending/ascending order",
                "S": "Show/hide last executed query for sleeping thread (Performance Schema only)",
                "u": "List users (results vary depending on if userstat variable is enabled)",
                "v": "Variable wildcard search via SHOW VARIABLES",
                "x": "Set the general refresh interval",
                "X": (
                    "Set the refresh interval for data the query SHOW ENGINE INNODB STATUS is responsible"
                    " for\n[grey70]This is good to change if your host has a very heavy workload since this query"
                    " can be a bottleneck"
                ),
                "y": "Display all databases on host",
                "z": "Show all entries in the host cache",
            }

            table_keys = Table(box=box.ROUNDED, style="grey70", title="Features", title_style="bold steel_blue1")
            table_keys.add_column("Key", justify="center", style="b steel_blue1")
            table_keys.add_column("Description")

            for key, description in sorted(keys.items()):
                table_keys.add_row("[steel_blue1]%s" % key, description, style=row_style)

            datapoints = {
                "Read Only": "If the host is in read-only mode",
                "Use PS": "If Dolphie is using Performance Schema for listing queries",
                "Read Hit": "The percentage of how many reads are from InnoDB buffer pool compared to from disk",
                "Lag": (
                    "Retrieves metric from: Slave -> SHOW SLAVE STATUS, HB -> Heartbeat table, PS -> Performance Schema"
                ),
                "Chkpt Age": (
                    "This depicts how close InnoDB is before it starts to furiously flush dirty data to disk "
                    "(Higher is better)"
                ),
                "AHI Hit": (
                    "The percentage of how many lookups there are from Adapative Hash Index compared to it not"
                    " being used"
                ),
                "Pending AIO": "W means writes, R means reads. The values should normally be 0",
                "Diff": "This is the size difference of the binary log between each refresh interval",
                "Cache Hit": "The percentage of how many binary log lookups are from cache instead of from disk",
                "History List": "History list length (number of un-purged row changes in InnoDB's undo logs)",
                "Unpurged TRX": (
                    "How many transactions are between the newest and the last purged in InnoDB's undo logs"
                ),
                "QPS": "Queries per second",
                "Latency": "How much time it takes to receive data from the host for Dolphie each refresh interval",
                "Threads": "Con = Connected, Run = Running, Cac = Cached from SHOW GLOBAL STATUS",
                "Speed": "How many seconds were taken off of replication lag from the last refresh interval",
                "Query A/Q": (
                    "How many queries are active/queued in InnoDB. Based on innodb_thread_concurrency variable"
                ),
                "Tickets": "Relates to innodb_concurrency_tickets variable",
            }

            table_info = Table(box=box.ROUNDED, style="grey70")
            table_info.add_column("Datapoint", style="steel_blue1")
            table_info.add_column("Description")
            for datapoint, description in sorted(datapoints.items()):
                table_info.add_row("[steel_blue1]%s" % datapoint, description, style=row_style)

            self.console.print(
                Align.center(self.base_title + " by Charles Thompson <[grey62]01charles.t@gmail.com[/grey62]>\n"),
                highlight=False,
            )

            table_grid = Table.grid()
            table_grid.add_row(table_panels, table_filters)
            self.console.print(Align.center(table_keys))
            self.console.print("")
            self.console.print(Align.center(table_grid))
            self.console.print("")
            self.console.print(Align.center(table_info))

            self.block_refresh_for_key_command()
        else:
            valid_key = False

        self.pause_refresh = False
        self.kb.set_new_term()

        return valid_key

    def create_user_stats_table(self):
        table = Table(header_style="bold white", box=box.ROUNDED, style="grey70")
        columns = {}

        if self.db.cursor.execute("SELECT @@userstat") == 1:
            userstat_enabled = self.db.cursor.fetchone()["@@userstat"] == 1

            if userstat_enabled:
                self.db.cursor.execute(Queries["userstat_user_statisitics"])
            elif self.performance_schema_enabled:
                self.db.cursor.execute(Queries["ps_user_statisitics"])
            else:
                return False

            users = self.db.cursor.fetchall()
            user_stats = {}

            for user in users:
                username = user["user"].decode()
                user_stats.setdefault(username, {}).update(
                    user=user["user"].decode(),
                    total_connections=user["total_connections"],
                    concurrent_connections=user.get("concurrent_connections"),
                    denied_connections=user.get("denied_connections"),
                    binlog_bytes_written=user.get("binlog_bytes_written"),
                    rows_fetched=user.get("rows_fetched"),
                    rows_updated=user.get("rows_updated"),
                    table_rows_read=user.get("table_rows_read"),
                    select_commands=user.get("select_commands"),
                    update_commands=user.get("update_commands"),
                    other_commands=user.get("other_commands"),
                    commit_transactions=user.get("commit_transactions"),
                    rollback_transactions=user.get("rollback_transactions"),
                    access_denied=user.get("access_denied"),
                    current_connections=user.get("current_connections"),
                    rows_affected=user.get("sum_rows_affected"),
                    rows_sent=user.get("sum_rows_sent"),
                    rows_read=user.get("sum_rows_examined"),
                    created_tmp_disk_tables=user.get("sum_created_tmp_disk_tables"),
                    created_tmp_tables=user.get("sum_created_tmp_tables"),
                )

            if userstat_enabled:
                columns.update(
                    {
                        "User": {"field": "user", "format_number": False},
                        "Active": {"field": "concurrent_connections", "format_number": True},
                        "Total": {"field": "total_connections", "format_number": True},
                        "Binlog Data": {"field": "binlog_bytes_written", "format_number": False},
                        "Rows Read": {"field": "table_rows_read", "format_number": True},
                        "Rows Sent": {"field": "rows_fetched", "format_number": True},
                        "Rows Updated": {"field": "rows_updated", "format_number": True},
                        "Selects": {"field": "select_commands", "format_number": True},
                        "Updates": {"field": "update_commands", "format_number": True},
                        "Other": {"field": "other_commands", "format_number": True},
                        "Commit": {"field": "commit_transactions", "format_number": True},
                        "Rollback": {"field": "rollback_transactions", "format_number": True},
                        "Access Denied": {"field": "access_denied", "format_number": True},
                        "Conn Denied": {"field": "denied_connections", "format_number": True},
                    }
                )
            else:
                columns.update(
                    {
                        "User": {"field": "user", "format_number": False},
                        "Active": {"field": "current_connections", "format_number": True},
                        "Total": {"field": "total_connections", "format_number": True},
                        "Rows Read": {"field": "rows_read", "format_number": True},
                        "Rows Sent": {"field": "rows_sent", "format_number": True},
                        "Rows Updated": {"field": "rows_affected", "format_number": True},
                        "Tmp Tables": {"field": "created_tmp_tables", "format_number": True},
                        "Tmp Disk Tables": {"field": "created_tmp_disk_tables", "format_number": True},
                    }
                )

            for column, data in columns.items():
                table.add_column(column, no_wrap=True)

            for user_data in user_stats.values():
                row_values = []
                for column, data in columns.items():
                    value = user_data.get(data["field"])
                    if column == "Binlog Data":
                        row_values.append(format_bytes(value) if value else "")
                    elif data["format_number"]:
                        row_values.append(format_number(value) if value else "")
                    else:
                        row_values.append(value or "")

                table.add_row(*row_values, style="grey93")

            return table if user_stats else False
        else:
            return False

    def load_host_cache_file(self):
        if os.path.exists(self.host_cache_file):
            with open(self.host_cache_file) as file:
                for line in file:
                    line = line.strip()
                    error_message = f"Host cache entry '{line}' is not properly formatted! Format: ip=hostname"

                    if "=" not in line:
                        raise Exception(error_message)

                    ip_address, hostname = line.split("=", maxsplit=1)
                    ip_address = ip_address.strip()
                    hostname = hostname.strip()

                    if not ip_address or not hostname:
                        raise Exception(error_message)

                    self.host_cache[ip_address] = hostname

    def get_hostname(self, ip_address):
        if ip_address in self.host_cache:
            return self.host_cache[ip_address]

        try:
            ipaddress.IPv4Network(ip_address)
            hostname = socket.gethostbyaddr(ip_address)[0]
            self.host_cache[ip_address] = hostname
        except (ValueError, socket.error):
            hostname = ip_address

        return hostname



#!/usr/bin/env python3

# ****************************
# *        Dolphie           *
# * Author: Charles Thompson *
# ****************************

import os
import re
from argparse import ArgumentParser, RawTextHelpFormatter
from configparser import ConfigParser
from datetime import datetime
from time import sleep

import myloginpath
from dolphie import Dolphie
from dolphie.Panels import (
    dashboard_panel,
    innodb_io_panel,
    innodb_locks_panel,
    processlist_panel,
    replica_panel,
)
from dolphie.Queries import Queries
from rich.live import Live
from rich.prompt import Prompt


def parse_args(dolphie: Dolphie):
    epilog = """
Config file with [client] section supports these options:
    host
    user
    password
    port
    socket
    ssl_mode REQUIRED/VERIFY_CA/VERIFY_IDENTITY
    ssl_ca
    ssl_cert
    ssl_key

Login path file supports these options:
    host
    user
    password
    port
    socket

Environment variables support these options:
    DOLPHIE_USER
    DOLPHIE_PASSWORD
    DOLPHIE_HOST
    DOLPHIE_PORT
    DOLPHIE_SOCKET

"""
    parser = ArgumentParser(
        conflict_handler="resolve",
        description="Dolphie, an intuitive feature-rich top tool for monitoring MySQL in real time",
        epilog=epilog,
        formatter_class=RawTextHelpFormatter,
    )

    parser.add_argument(
        "-u",
        "--user",
        dest="user",
        type=str,
        help="Username for MySQL",
    )
    parser.add_argument("-p", "--password", dest="password", type=str, help="Password for MySQL")
    parser.add_argument(
        "--ask-pass",
        dest="ask_password",
        action="store_true",
        default=False,
        help="Ask for password (hidden text)",
    )
    parser.add_argument(
        "-h",
        "--host",
        dest="host",
        type=str,
        help="Hostname/IP address for MySQL",
    )
    parser.add_argument(
        "-P",
        "--port",
        dest="port",
        type=int,
        help="Port for MySQL (Socket has precendence)",
    )
    parser.add_argument(
        "-S",
        "--socket",
        dest="socket",
        type=str,
        help="Socket file for MySQL",
    )
    parser.add_argument(
        "-c",
        "--config-file",
        dest="config_file",
        type=str,
        help=(
            "Absolute config file path to use. This should use [client] section. "
            "See below for options support [default: ~/.my.cnf]"
        ),
    )
    parser.add_argument(
        "-f",
        "--host-cache-file",
        dest="host_cache_file",
        type=str,
        help=(
            "Resolve IPs to hostnames when your DNS is unable to. Each IP/hostname pair should be on its own line "
            "using format: ip=hostname [default: %(default)s]"
        ),
    )
    parser.add_argument(
        "-l",
        "--login-path",
        dest="login_path",
        default="client",
        type=str,
        help=(
            "Specify login path to use mysql_config_editor's file ~/.mylogin.cnf for encrypted login credentials. "
            "Supercedes config file [default: %(default)s]"
        ),
    )
    parser.add_argument(
        "-r",
        "--refresh_interval",
        dest="refresh_interval",
        default=1,
        type=int,
        help="How much time to wait in seconds between each refresh [default: %(default)s]",
    )
    parser.add_argument(
        "-R",
        "--refresh_interval_innodb_status",
        dest="refresh_interval_innodb_status",
        default=1,
        type=int,
        help=(
            "How much time to wait in seconds to execute SHOW ENGINE INNODB STATUS to refresh data its responsible "
            "for [default: %(default)s]"
        ),
    )
    parser.add_argument(
        "-H",
        "--heartbeat-table",
        dest="heartbeat_table",
        type=str,
        help=(
            "If your hosts use pt-heartbeat, specify table in format db.table to use the timestamp it "
            "has for replication lag instead of Seconds_Behind_Master from SHOW SLAVE STATUS"
        ),
    )
    parser.add_argument(
        "--ssl-mode",
        dest="ssl_mode",
        type=str,
        help=(
            "Desired security state of the connection to the host. Supports: "
            "REQUIRED/VERIFY_CA/VERIFY_IDENTITY [default: OFF]"
        ),
    )
    parser.add_argument(
        "--ssl-ca",
        dest="ssl_ca",
        type=str,
        help="Path to the file that contains a PEM-formatted CA certificate",
    )
    parser.add_argument(
        "--ssl-cert",
        dest="ssl_cert",
        type=str,
        help="Path to the file that contains a PEM-formatted client certificate",
    )
    parser.add_argument(
        "--ssl-key",
        dest="ssl_key",
        type=str,
        help="Path to the file that contains a PEM-formatted private key for the client certificate",
    )
    parser.add_argument(
        "--hide-dashboard",
        dest="dashboard",
        action="store_false",
        default=True,
        help=(
            "Start without showing dashboard. This is good to use if you want to reclaim terminal space and "
            "not execute the additional queries for it"
        ),
    )
    parser.add_argument(
        "--show-trxs-only",
        dest="show_trxs_only",
        action="store_true",
        default=False,
        help="Start with only showing queries that are running a transaction",
    )
    parser.add_argument(
        "--additional-columns",
        dest="show_additional_query_columns",
        action="store_true",
        default=False,
        help="Start with additional columns in processlist panel",
    )
    parser.add_argument(
        "--use-processlist",
        dest="use_processlist",
        action="store_true",
        default=False,
        help="Start with using Processlist instead of Performance Schema for listing queries",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        default=False,
        help="Print tracebacks on errors for more verbose debugging",
    )
    parser.add_argument(
        "-V", "--version", action="version", version=dolphie.app_version, help="Display version and exit"
    )

    parameter_options = vars(parser.parse_args())  # Convert object to dict
    basic_options = ["user", "password", "host", "port", "socket"]

    # Use specified config file if there is one, else use standard ~/.my.cnf
    if parameter_options["config_file"]:
        dolphie.config_file = parameter_options["config_file"]
    else:
        dolphie.config_file = "%s/.my.cnf" % os.path.expanduser("~")

    if parameter_options["debug"]:
        dolphie.debug = parameter_options["debug"]

    # Use config file for login credentials
    if os.path.isfile(dolphie.config_file):
        cfg = ConfigParser()
        cfg.read(dolphie.config_file)

        for option in basic_options:
            if cfg.has_option("client", option):
                setattr(dolphie, option, cfg.get("client", option))

        if cfg.has_option("client", "ssl_mode"):
            ssl_mode = cfg.get("client", "ssl_mode").upper()

            if ssl_mode == "REQUIRED":
                dolphie.ssl[""] = True
            elif ssl_mode == "VERIFY_CA":
                dolphie.ssl["check_hostname"] = False
            elif ssl_mode == "VERIFY_IDENTITY":
                dolphie.ssl["check_hostname"] = True
            else:
                raise Exception("Unsupported SSL mode [b]%s" % ssl_mode)

        if cfg.has_option("client", "ssl_ca"):
            dolphie.ssl["ca"] = cfg.get("client", "ssl_ca")
        if cfg.has_option("client", "ssl_cert"):
            dolphie.ssl["cert"] = cfg.get("client", "ssl_cert")
        if cfg.has_option("client", "ssl_key"):
            dolphie.ssl["key"] = cfg.get("client", "ssl_key")

    # Use login path for login credentials
    if parameter_options["login_path"]:
        try:
            login_path_data = myloginpath.parse(parameter_options["login_path"])

            for option in basic_options:
                if option in login_path_data:
                    setattr(dolphie, option, login_path_data[option])
        except Exception as e:
            # Don't error out for default login path
            if parameter_options["login_path"] != "client":
                raise Exception(f"Problem reading login path file Reason: {e}")

    # Use environment variables for basic options if specified
    for option in basic_options:
        environment_var = "DOLPHIE_%s" % option.upper()
        if environment_var in os.environ and os.environ[environment_var]:
            setattr(dolphie, option, os.environ[environment_var])

    # Lastly, use parameter options if specified
    for option in basic_options:
        if parameter_options[option]:
            setattr(dolphie, option, parameter_options[option])

    if parameter_options["ask_password"]:
        dolphie.password = Prompt.ask("[b steel_blue1]Password", password=True)

    if not dolphie.host:
        dolphie.host = "localhost"

    if parameter_options["refresh_interval"]:
        dolphie.refresh_interval = parameter_options["refresh_interval"]

    if parameter_options["refresh_interval_innodb_status"]:
        dolphie.refresh_interval_innodb_status = parameter_options["refresh_interval_innodb_status"]

    if parameter_options["heartbeat_table"]:
        pattern_match = re.search(r"^(\w+\.\w+)$", parameter_options["heartbeat_table"])
        if pattern_match:
            dolphie.heartbeat_table = parameter_options["heartbeat_table"]
            Queries["heartbeat_replica_lag"] = Queries["heartbeat_replica_lag"].replace(
                "$placeholder", dolphie.heartbeat_table
            )
        else:
            raise Exception("Your heartbeat table did not conform to the proper format db.table")

    if parameter_options["ssl_mode"]:
        ssl_mode = parameter_options["ssl_mode"].upper()

        if ssl_mode == "REQUIRED":
            dolphie.ssl[""] = True
        elif ssl_mode == "VERIFY_CA":
            dolphie.ssl["check_hostame"] = False
        elif ssl_mode == "VERIFY_IDENTITY":
            dolphie.ssl["check_hostame"] = True
        else:
            raise Exception(f"Unsupported SSL mode {ssl_mode}")

    if parameter_options["ssl_ca"]:
        dolphie.ssl["ca"] = parameter_options["ssl_ca"]
    if parameter_options["ssl_cert"]:
        dolphie.ssl["cert"] = parameter_options["ssl_cert"]
    if parameter_options["ssl_key"]:
        dolphie.ssl["key"] = parameter_options["ssl_key"]

    if parameter_options["host_cache_file"]:
        dolphie.host_cache_file = parameter_options["host_cache_file"]
    else:
        dolphie.host_cache_file = os.path.dirname(os.path.abspath(__file__)) + "/host_cache"

    dolphie.dashboard = parameter_options["dashboard"]
    dolphie.show_trxs_only = parameter_options["show_trxs_only"]
    dolphie.show_additional_query_columns = parameter_options["show_additional_query_columns"]
    dolphie.use_processlist = parameter_options["use_processlist"]

    if dolphie.dashboard:
        dolphie.layout["dashboard"].visible = True
    else:
        dolphie.layout["dashboard"].visible = False


def main():
    dolphie = Dolphie()

    try:
        dolphie.create_rich_layout()
        parse_args(dolphie)
        dolphie.check_for_update()
        dolphie.db_connect()
        dolphie.load_host_cache_file()

        with Live(
            dolphie.layout, vertical_overflow="crop", screen=True, transient=True, auto_refresh=False
        ) as dolphie.rich_live:
            while True:
                if dolphie.pause_refresh is False:
                    loop_time = datetime.now()

                    dolphie.statuses = dolphie.fetch_data("status")
                    if dolphie.first_loop:
                        dolphie.saved_status = dolphie.statuses.copy()

                    dolphie.loop_duration_seconds = (loop_time - dolphie.previous_main_loop_time).total_seconds()
                    loop_duration_innodb_status_seconds = (
                        loop_time - dolphie.previous_innodb_status_loop_time
                    ).total_seconds()

                    if dolphie.layout["processlist"].visible:
                        dolphie.processlist_threads = processlist_panel.get_data(dolphie)
                        dolphie.layout["processlist"].update(processlist_panel.create_panel(dolphie))

                    if dolphie.dashboard:
                        dolphie.variables = dolphie.fetch_data("variables")
                        dolphie.primary_status = dolphie.fetch_data("primary_status")
                        dolphie.replica_status = dolphie.fetch_data("replica_status")

                        if (
                            dolphie.first_loop
                            or loop_duration_innodb_status_seconds >= dolphie.refresh_interval_innodb_status
                        ):
                            dolphie.innodb_status = dolphie.fetch_data("innodb_status")

                        dolphie.layout["dashboard"].update(dashboard_panel.create_panel(dolphie))

                        # Save some variables to be used in next refresh
                        dolphie.previous_binlog_position = 0
                        if dolphie.primary_status:
                            dolphie.previous_binlog_position = dolphie.primary_status["Position"]

                    if dolphie.layout["replicas"].visible:
                        dolphie.layout["replicas"].update(replica_panel.create_panel(dolphie))

                    if dolphie.layout["innodb_io"].visible:
                        dolphie.layout["innodb_io"].update(innodb_io_panel.create_panel(dolphie))

                    if dolphie.layout["innodb_locks"].visible:
                        dolphie.layout["innodb_locks"].update(innodb_locks_panel.create_panel(dolphie))

                    if dolphie.replica_status:
                        dolphie.previous_replica_sbm = 0
                        if dolphie.replica_status["Seconds_Behind_Master"] is not None:
                            dolphie.previous_replica_sbm = dolphie.replica_status["Seconds_Behind_Master"]

                    # This is for the many stats per second in Dolphie
                    dolphie.saved_status = dolphie.statuses.copy()
                    dolphie.previous_main_loop_time = loop_time

                    if loop_duration_innodb_status_seconds >= dolphie.refresh_interval_innodb_status:
                        dolphie.previous_innodb_status_loop_time = loop_time

                    dolphie.rich_live.update(dolphie.layout, refresh=True)
                else:
                    # To prevent main loop from eating up 100% CPU
                    sleep(0.01)

                # Detect a keypress loop
                loop_counter = 0
                while loop_counter <= dolphie.refresh_interval * 10:
                    if dolphie.kb.key_press():
                        if dolphie.pause_refresh is False:
                            key = dolphie.kb.getch()

                        if key:
                            valid_key = dolphie.capture_key(key)
                            if valid_key:
                                break

                    # refresh_interval * 10 * 100ms equals to the second of sleep a user wants. This allows us to
                    # capture a key faster than sleeping the whole second of refresh_interval
                    sleep(0.100)

                    loop_counter += 1

                dolphie.first_loop = False
    except Exception as e:
        if dolphie.debug:
            dolphie.console.print_exception()
        else:
            dolphie.console.print(f"[b][red]ERROR[/red]![/b] {str(e)}", highlight=False)


if __name__ == "__main__":
    main()



import re

from dolphie import Dolphie
from dolphie.Functions import format_bytes, format_number
from rich import box
from rich.align import Align
from rich.console import Group
from rich.style import Style
from rich.table import Table


def create_panel(dolphie: Dolphie):
    variables = dolphie.variables
    statuses = dolphie.statuses
    loop_duration_seconds = dolphie.loop_duration_seconds
    saved_status = dolphie.saved_status

    row_style = Style(color="grey93")
    table_title_style = Style(color="grey93", bold=True)

    # Only run this if dashboard isn't turned on
    innodb_status = dolphie.innodb_status
    if dolphie.dashboard is False:
        innodb_status = dolphie.fetch_data("innodb_status")

    table_innodb_information = Table(
        box=box.ROUNDED,
        style="grey70",
        title="InnoDB Information",
        title_style=table_title_style,
        show_header=False,
    )
    table_innodb_information.add_column("")
    table_innodb_information.add_column("")

    table_innodb_information.add_row(
        "[grey78]BP Size",
        format_bytes(float(variables["innodb_buffer_pool_size"])),
        style=row_style,
    )
    table_innodb_information.add_row(
        "[grey78]BP Available",
        format_bytes(float(variables["innodb_buffer_pool_size"]) - float(statuses["Innodb_buffer_pool_bytes_data"])),
        style=row_style,
    )
    table_innodb_information.add_row(
        "[grey78]BP Dirty",
        format_bytes(float(statuses["Innodb_buffer_pool_bytes_dirty"])),
        style=row_style,
    )
    # MariaDB Support
    if "innodb_buffer_pool_instances" in variables:
        bp_instances = str(variables["innodb_buffer_pool_instances"])
    else:
        bp_instances = 1
    table_innodb_information.add_row("[grey78]BP Instances", str(bp_instances), style=row_style)
    table_innodb_information.add_row(
        "[grey78]BP Pages Free",
        format_number(float(statuses["Innodb_buffer_pool_pages_free"])),
        style=row_style,
    )

    # MariaDB Support
    if "innodb_log_files_in_group" in variables:
        log_files_in_group = variables["innodb_log_files_in_group"]
    else:
        log_files_in_group = 1

    table_innodb_information.add_row(
        "[grey78]Total Log Size",
        format_bytes(variables["innodb_log_file_size"] * log_files_in_group),
        style=row_style,
    )

    if "innodb_adaptive_hash_index_parts" in variables.keys():
        table_innodb_information.add_row(
            "[grey78]Adapt Hash Idx",
            "%s [grey78]([grey93]%s[grey78])"
            % (variables["innodb_adaptive_hash_index"], variables["innodb_adaptive_hash_index_parts"]),
            style=row_style,
        )
    else:
        table_innodb_information.add_row(
            "[grey78]Adapt Hash Idx",
            "%s [grey78]" % (variables["innodb_adaptive_hash_index"]),
            style=row_style,
        )

    # Calculate AIO reads
    total_pending_aio_reads = 0
    total_pending_aio_writes = 0
    output = re.search(
        r"Pending normal aio reads: (?:\d+\s)?\[(.*?)\] , aio writes: (?:\d+\s)?\[(.*?)\]", innodb_status["status"]
    )
    if output:
        match = output.group(1).split(",")

        for aio_read in match:
            total_pending_aio_reads += int(aio_read)

        match = output.group(2).split(",")
        for aio_write in match:
            total_pending_aio_writes += int(aio_write)
    else:
        total_pending_aio_reads = "N/A"
        total_pending_aio_writes = "N/A"

    # Get pending insert buffer, log i/o, and sync i/o
    pending_ibuf_aio_reads = 0
    pending_log_ios = 0
    pending_sync_ios = 0

    search_pattern = re.search(
        r"ibuf aio reads:\s?(\d+)?, log i/o's:\s?(\d+)?, sync i/o's:\s?(\d+)",
        innodb_status["status"],
    )
    if search_pattern:
        pending_ibuf_aio_reads = search_pattern.group(1)
        pending_log_ios = search_pattern.group(2)
        pending_sync_ios = search_pattern.group(3)

    # Get pending log and buffer pool flushes
    pending_log_flush = 0
    pending_buffer_pool_flush = 0

    search_pattern = re.search(r"Pending flushes \(.+?\) log: (\d+); buffer pool: (\d+)", innodb_status["status"])
    if search_pattern:
        pending_log_flush = search_pattern.group(1)
        pending_buffer_pool_flush = search_pattern.group(2)

    table_pending_io = Table(
        box=box.ROUNDED,
        style="grey70",
        title="Pending",
        title_style=table_title_style,
        show_header=False,
    )
    table_pending_io.add_column("")
    table_pending_io.add_column("", min_width=5)

    table_pending_io.add_row(
        "[grey78]Normal AIO Reads",
        format_number(total_pending_aio_reads),
        style=row_style,
    )
    table_pending_io.add_row(
        "[grey78]Normal AIO Writes",
        format_number(total_pending_aio_writes),
        style=row_style,
    )
    table_pending_io.add_row(
        "[grey78]Insert Buffer Reads",
        format_number(pending_ibuf_aio_reads),
        style=row_style,
    )
    table_pending_io.add_row("[grey78]Log IO/s", format_number(pending_log_ios), style=row_style)
    table_pending_io.add_row("[grey78]Sync IO/s", format_number(pending_sync_ios), style=row_style)
    table_pending_io.add_row("[grey78]Log Flushes", format_number(pending_log_flush), style=row_style)
    table_pending_io.add_row(
        "[grey78]Buffer Pool Flushes",
        format_number(pending_buffer_pool_flush),
        style=row_style,
    )

    # Get reads/avg bytes/writes/fsyncs per second
    reads_s = 0
    writes_s = 0
    fsyncs_s = 0
    bytes_s = 0

    search_pattern = re.search(
        r"(\d+\.?\d*) reads/s, (\d+\.?\d*) avg bytes/read, (\d+\.?\d*) writes/s, (\d+\.?\d*) fsyncs/s",
        innodb_status["status"],
    )
    if search_pattern:
        reads_s = search_pattern.group(1)
        writes_s = search_pattern.group(3)
        fsyncs_s = search_pattern.group(4)
        bytes_s = search_pattern.group(2)

    # Get OS file reads, OS file writes, and OS fsyncs
    os_file_reads = 0
    os_file_writes = 0
    os_fsyncs = 0

    search_pattern = re.search(
        r"(\d+) OS file reads, (\d+) OS file writes, (\d+) OS fsyncs",
        innodb_status["status"],
    )
    if search_pattern:
        os_file_reads = search_pattern.group(1)
        os_file_writes = search_pattern.group(2)
        os_fsyncs = search_pattern.group(3)

    table_file_io = Table(
        box=box.ROUNDED,
        style="grey70",
        title="File",
        title_style=table_title_style,
        show_header=False,
    )
    table_file_io.add_column("")
    table_file_io.add_column("", min_width=6)

    table_file_io.add_row("[grey78]OS Reads", format_number(os_file_reads), style=row_style)
    table_file_io.add_row("[grey78]OS Writes", format_number(os_file_writes), style=row_style)
    table_file_io.add_row("[grey78]OS fsyncs", format_number(os_fsyncs), style=row_style)
    table_file_io.add_row("[grey78]Read/s", format_number(reads_s), style=row_style)
    table_file_io.add_row("[grey78]Write/s", format_number(writes_s), style=row_style)
    table_file_io.add_row("[grey78]FSync/s", format_number(fsyncs_s), style=row_style)
    table_file_io.add_row("[grey78]Bytes/s", format_number(bytes_s), style=row_style)

    table_innodb_activity = Table(
        box=box.ROUNDED,
        style="grey70",
        title="Activity",
        title_style=table_title_style,
        show_header=False,
    )
    table_innodb_activity.add_column("")
    table_innodb_activity.add_column("", width=7)

    if loop_duration_seconds == 0:
        reads_mem_per_second = 0
        reads_disk_per_second = 0
        writes_per_second = 0
        log_waits = 0
        row_lock_waits = 0
    else:
        reads_mem_per_second = round(
            (statuses["Innodb_buffer_pool_read_requests"] - saved_status["Innodb_buffer_pool_read_requests"])
            / loop_duration_seconds
        )
        reads_disk_per_second = round(
            (statuses["Innodb_buffer_pool_reads"] - saved_status["Innodb_buffer_pool_reads"]) / loop_duration_seconds
        )
        writes_per_second = round(
            (statuses["Innodb_buffer_pool_write_requests"] - saved_status["Innodb_buffer_pool_write_requests"])
            / loop_duration_seconds
        )

        log_waits = round((statuses["Innodb_log_waits"] - saved_status["Innodb_log_waits"]) / loop_duration_seconds)

        row_lock_waits = round(
            (statuses["Innodb_row_lock_waits"] - saved_status["Innodb_row_lock_waits"]) / loop_duration_seconds
        )

    table_innodb_activity.add_row("[grey78]BP reads/s (mem)", format_number(reads_mem_per_second), style=row_style)
    table_innodb_activity.add_row("[grey78]BP reads/s (disk)", format_number(reads_disk_per_second), style=row_style)
    table_innodb_activity.add_row("[grey78]BP writes/s", format_number(writes_per_second), style=row_style)

    bp_clean_page_wait = format_number(
        (statuses["Innodb_buffer_pool_wait_free"] - saved_status["Innodb_buffer_pool_wait_free"])
        / loop_duration_seconds
    )

    bp_clean_page_wait_color = ""
    if bp_clean_page_wait != "0":
        bp_clean_page_wait_color = "[bright_red]"

    table_innodb_activity.add_row(
        "[grey78]BP clean page wait/s",
        bp_clean_page_wait_color + bp_clean_page_wait,
        style=row_style,
    )
    table_innodb_activity.add_row("[grey78]Log waits/s", format_number(log_waits), style=row_style)
    table_innodb_activity.add_row("[grey78]Row lock waits/s", format_number(row_lock_waits), style=row_style)
    table_innodb_activity.add_row(
        "[grey78]Row lock time avg", "%sms" % str(statuses["Innodb_row_lock_time_avg"]), style=row_style
    )

    table_row_operations = Table(
        box=box.ROUNDED,
        style="grey70",
        title="Row Operations/s",
        title_style=table_title_style,
        show_header=False,
    )
    table_row_operations.add_column("")
    table_row_operations.add_column("", min_width=8)

    if loop_duration_seconds == 0:
        reads_per_second = 0
        inserts_per_second = 0
        updates_per_second = 0
        deletes_per_second = 0
    else:
        reads_per_second = round(
            (statuses["Innodb_rows_read"] - saved_status["Innodb_rows_read"]) / loop_duration_seconds
        )
        inserts_per_second = round(
            (statuses["Innodb_rows_inserted"] - saved_status["Innodb_rows_inserted"]) / loop_duration_seconds
        )
        updates_per_second = round(
            (statuses["Innodb_rows_updated"] - saved_status["Innodb_rows_updated"]) / loop_duration_seconds
        )
        deletes_per_second = round(
            (statuses["Innodb_rows_deleted"] - saved_status["Innodb_rows_deleted"]) / loop_duration_seconds
        )

    table_row_operations.add_row("[grey78]Reads", format_number(reads_per_second))
    table_row_operations.add_row("[grey78]Inserts", format_number(inserts_per_second))
    table_row_operations.add_row("[grey78]Updates", format_number(updates_per_second))
    table_row_operations.add_row("[grey78]Deletes", format_number(deletes_per_second))
    table_row_operations.add_row()
    table_row_operations.add_row()
    table_row_operations.add_row()

    # Put these two tables side-by-side
    table_grid = Table.grid()
    table_grid.add_row(
        table_innodb_information, table_innodb_activity, table_row_operations, table_pending_io, table_file_io
    )

    return Group(Align.center(table_grid))



import re
from datetime import timedelta

import pymysql
from dolphie import Dolphie
from dolphie.Queries import Queries
from rich import box
from rich.align import Align
from rich.console import Group
from rich.style import Style
from rich.table import Table
from rich.text import Text


def create_panel(dolphie: Dolphie):
    table_grid = Table.grid()
    table_replication = Table()

    # Only run this if dashboard isn't turned on
    if dolphie.dashboard is not True:
        dolphie.replica_status = dolphie.fetch_data("replica_status")

    if dolphie.replica_status:
        table_replication = create_table(dolphie, dolphie.replica_status)

    find_replicas_query = 0
    if dolphie.use_performance_schema:
        find_replicas_query = Queries["ps_find_replicas"]
    else:
        find_replicas_query = Queries["pl_find_replicas"]

    replica_count = dolphie.db.cursor.execute(find_replicas_query)
    data = dolphie.db.cursor.fetchall()

    replica_count_text = Text.from_markup("\n[b steel_blue1]%s[/b steel_blue1] replicas" % (replica_count))

    table_split_counter = 1
    tables = []
    for row in data:
        thread_id = row["id"]

        # Resolve IPs to addresses and add to cache for fast lookup
        host = dolphie.get_hostname(row["host"].decode().split(":")[0])

        table = Table(box=box.ROUNDED, show_header=False, style="grey70")
        try:
            if thread_id not in dolphie.replica_connections:
                dolphie.replica_connections[thread_id] = {
                    "host": host,
                    "connection": pymysql.connect(
                        host=host,
                        user=dolphie.user,
                        passwd=dolphie.password,
                        port=dolphie.port,
                        ssl=dolphie.ssl,
                        autocommit=True,
                    ),
                    "previous_sbm": 0,
                }

            replica_connection = dolphie.replica_connections[thread_id]
            replica_cursor = replica_connection["connection"].cursor(pymysql.cursors.DictCursor)
            replica_cursor.execute(Queries["replica_status"])
            replica_data = replica_cursor.fetchone()

            if replica_data:
                tables.append(create_table(dolphie, replica_data, list_replica_thread_id=thread_id))
        except pymysql.Error as e:
            row_style = Style(color="grey93")

            table.add_column()
            table.add_column(width=30)

            table.add_row("[grey78]Host", host, style=row_style)
            table.add_row("[grey78]User", row["user"].decode(), style=row_style)
            table.add_row("[bright_red]Error", e.args[1], style=row_style)

            tables.append(table)

        if table_split_counter == 3:
            table_grid.add_row(*tables)
            table_split_counter = 0
            tables = []

        table_split_counter += 1

    if table_split_counter:
        table_grid.add_row(*tables)

    if dolphie.replica_status:
        # GTID Sets can be very long, so we don't center align replication table or else table
        # will increase/decrease in size a lot
        if ("Executed_Gtid_Set" in dolphie.replica_status and dolphie.replica_status["Executed_Gtid_Set"]) or (
            "Using_Gtid" in dolphie.replica_status and dolphie.replica_status["Using_Gtid"] != "No"
        ):
            panel_data = Group(
                Align.left(table_replication),
                Align.center(replica_count_text),
                Align.center(table_grid),
            )
        else:
            panel_data = Group(
                Align.center(table_replication),
                Align.center(replica_count_text),
                Align.center(table_grid),
            )
    else:
        panel_data = Group(Align.center(replica_count_text), Align.center(table_grid))

    return panel_data


def create_table(dolphie: Dolphie, data, dashboard_table=False, list_replica_thread_id=None):
    # This is for the replica view
    if list_replica_thread_id:
        if dolphie.replica_connections[list_replica_thread_id]["previous_sbm"] is not None:
            replica_previous_replica_sbm = dolphie.replica_connections[list_replica_thread_id]["previous_sbm"]
    else:
        replica_previous_replica_sbm = dolphie.previous_replica_sbm

    if data["Slave_IO_Running"].lower() == "no":
        data["Slave_IO_Running"] = "[bright_red]NO[/bright_red]"
    else:
        data["Slave_IO_Running"] = "[bright_green]Yes[/bright_green]"

    if data["Slave_SQL_Running"].lower() == "no":
        data["Slave_SQL_Running"] = "[bright_red]NO[/bright_red]"
    else:
        data["Slave_SQL_Running"] = "[bright_green]Yes[/bright_green]"

    data["sbm_source"] = "Replica"
    # Use performance schema for seconds behind if host is MySQL 8
    if dolphie.full_version.startswith("8") and dolphie.performance_schema_enabled:
        dolphie.db.cursor.execute(Queries["ps_replica_lag"])
        replica_lag_data = dolphie.db.cursor.fetchone()

        data["sbm_source"] = "PS"
        if replica_lag_data["secs_behind"]:
            data["Seconds_Behind_Master"] = int(replica_lag_data["secs_behind"])
        else:
            data["Seconds_Behind_Master"] = 0
    # Use heartbeat table from pt-toolkit if specified
    elif dolphie.heartbeat_table:
        try:
            dolphie.db.cursor.execute(Queries["heartbeat_replica_lag"])
            replica_lag_data = dolphie.db.cursor.fetchone()

            if replica_lag_data["secs_behind"] is not None:
                data["sbm_source"] = "HB"
                data["Seconds_Behind_Master"] = int(replica_lag_data["secs_behind"])
        except pymysql.Error:
            pass

    data["speed"] = 0
    # Colorize seconds behind
    if data["Seconds_Behind_Master"] is not None:
        replica_sbm = data["Seconds_Behind_Master"]

        if list_replica_thread_id:
            dolphie.replica_connections[list_replica_thread_id]["previous_sbm"] = replica_sbm

        if replica_previous_replica_sbm and replica_sbm < replica_previous_replica_sbm:
            data["speed"] = round((replica_previous_replica_sbm - replica_sbm) / dolphie.refresh_interval)

        if replica_sbm != 0:
            if replica_sbm > 20:
                data["lag"] = "[bright_red]%s" % "{:0>8}".format(str(timedelta(seconds=replica_sbm)))
            elif replica_sbm > 10:
                data["lag"] = "[bright_yellow]%s" % "{:0>8}".format(str(timedelta(seconds=replica_sbm)))
            else:
                data["lag"] = "[bright_green]%s" % "{:0>8}".format(str(timedelta(seconds=replica_sbm)))
        elif replica_sbm == 0:
            data["lag"] = "[bright_green]00:00:00"

    data["Master_Host"] = dolphie.get_hostname(data["Master_Host"])
    data["mysql_gtid_enabled"] = False
    data["mariadb_gtid_enabled"] = False
    data["gtid"] = "OFF"
    if "Executed_Gtid_Set" in data and data["Executed_Gtid_Set"]:
        data["mysql_gtid_enabled"] = True
        data["gtid"] = "ON"
    if "Using_Gtid" in data and data["Using_Gtid"] != "No":
        data["mariadb_gtid_enabled"] = True
        data["gtid"] = data["Using_Gtid"]

    # Create table for Rich
    table_title_style = Style(color="grey93", bold=True)
    table_line_color = "grey78"
    row_style = Style(color="grey93")

    table_title = ""
    if dashboard_table is True or list_replica_thread_id is None:
        table_title = "Replication"

    table = Table(
        show_header=False,
        box=box.ROUNDED,
        title=table_title,
        title_style=table_title_style,
        style=table_line_color,
    )

    table.add_column()
    if dashboard_table is True:
        table.add_column(max_width=21)
    elif list_replica_thread_id is not None:
        if data["mysql_gtid_enabled"] or data["mariadb_gtid_enabled"]:
            table.add_column(max_width=60)
    else:
        table.add_column(overflow="fold")

    if list_replica_thread_id is not None:
        table.add_row(
            "[grey78]Host", "[grey93]%s" % dolphie.replica_connections[list_replica_thread_id]["host"], style=row_style
        )
    else:
        table.add_row("[grey78]Primary", "[grey93]%s" % data["Master_Host"], style=row_style)

    table.add_row("[grey78]User", "[grey93]%s" % data["Master_User"], style=row_style)
    table.add_row(
        "[grey78]Thread",
        "[grey78]IO [grey93]%s [grey78]SQL [grey93]%s" % (data["Slave_IO_Running"], data["Slave_SQL_Running"]),
        style=row_style,
    )
    if data["Seconds_Behind_Master"] is None:
        table.add_row("[grey78]Lag", "", style=row_style)
    else:
        table.add_row(
            "[grey78]%s Lag" % data["sbm_source"],
            "[grey93]%s [grey78]Speed [grey93]%s" % (data["lag"], data["speed"]),
            style=row_style,
        )
    if dashboard_table:
        table.add_row("[grey78]Binlog IO", "[grey93]%s" % (data["Master_Log_File"]), style=row_style)
        table.add_row("[grey78]Binlog SQL", "[grey93]%s" % (data["Relay_Master_Log_File"]), style=row_style)
        table.add_row("[grey78]Relay Log ", "[grey93]%s" % (data["Relay_Log_File"]), style=row_style)
        table.add_row("[grey78]GTID", "[grey93]%s" % data["gtid"], style=row_style)
    else:
        table.add_row(
            "[grey78]Binlog IO",
            "%s ([grey62]%s[/grey62])" % (data["Master_Log_File"], data["Read_Master_Log_Pos"]),
            style=row_style,
        )
        table.add_row(
            "[grey78]Binlog SQL",
            "%s ([grey62]%s[/grey62])"
            % (
                data["Relay_Master_Log_File"],
                data["Exec_Master_Log_Pos"],
            ),
            style=row_style,
        )
        table.add_row(
            "[grey78]Relay Log",
            "%s ([grey62]%s[/grey62])" % (data["Relay_Log_File"], data["Relay_Log_Pos"]),
            style=row_style,
        )

        table.add_row("[grey78]GTID", "[grey93]%s" % data["gtid"], style=row_style)
        if data["mysql_gtid_enabled"]:
            executed_gtid_set = data["Executed_Gtid_Set"]
            retrieved_gtid_set = data["Retrieved_Gtid_Set"]

            for m in re.findall(r"(\w+-\w+-\w+-\w+-)(\w+)", retrieved_gtid_set):
                source_id = m[0] + m[1]
                if source_id == dolphie.server_uuid:
                    retrieved_gtid_set = retrieved_gtid_set.replace(m[0], "[grey62]…[medium_spring_green]")
                else:
                    retrieved_gtid_set = retrieved_gtid_set.replace(m[0], "[grey62]…[steel_blue1]")

                retrieved_gtid_set = retrieved_gtid_set.replace(m[1], "%s[grey93]" % m[1])

            for m in re.findall(r"(\w+-\w+-\w+-\w+-)(\w+)", executed_gtid_set):
                source_id = m[0] + m[1]
                if source_id == dolphie.server_uuid:
                    executed_gtid_set = executed_gtid_set.replace(m[0], "[grey62]…[medium_spring_green]")
                else:
                    executed_gtid_set = executed_gtid_set.replace(m[0], "[grey62]…[steel_blue1]")

                executed_gtid_set = executed_gtid_set.replace(m[1], "%s[grey93]" % m[1])

            table.add_row("[grey78]Auto Position", "[grey93]%s" % data["Auto_Position"], style=row_style)
            table.add_row("[grey78]Retrieved GTID Set", "[grey93]%s" % retrieved_gtid_set, style=row_style)
            table.add_row("[grey78]Executed GTID Set", "[grey93]%s" % executed_gtid_set, style=row_style)
        elif data["mariadb_gtid_enabled"]:
            table.add_row("[grey78]GTID IO Pos ", "[grey93]%s" % data["Gtid_IO_Pos"], style=row_style)

        if data["Last_Error"]:
            table.add_row("[grey78]Error ", "[grey93]%s" % data["Last_Error"])
        elif data["Last_IO_Error"]:
            table.add_row("[grey78]Error ", "[grey93]%s" % data["Last_IO_Error"])
        elif data["Last_SQL_Error"]:
            table.add_row("[grey78]Error ", "[grey93]%s" % data["Last_SQL_Error"])

    return table



import re
from datetime import datetime, timedelta

from dolphie import Dolphie
from dolphie.Functions import format_bytes, format_number
from dolphie.Panels import replica_panel
from rich import box
from rich.align import Align
from rich.panel import Panel
from rich.style import Style
from rich.table import Table


def create_panel(dolphie: Dolphie):
    statuses = dolphie.statuses
    variables = dolphie.variables
    innodb_status = dolphie.innodb_status
    saved_status = dolphie.saved_status
    loop_duration_seconds = dolphie.loop_duration_seconds
    primary_status = dolphie.primary_status

    tables_to_add = []
    uptime = str(timedelta(seconds=statuses["Uptime"]))

    dashboard_grid = Table.grid()
    dashboard_grid.add_column()
    dashboard_grid.add_column()
    dashboard_grid.add_column()

    row_style = Style(color="gray78")
    table_title_style = Style(color="grey93", bold=True)
    table_box = box.ROUNDED
    table_line_color = "grey78"

    ################
    # Information #
    ###############
    table_information = Table(
        show_header=False,
        box=box.ROUNDED,
        title="Host Information",
        title_style=table_title_style,
        style=table_line_color,
    )

    if loop_duration_seconds < 1:
        refresh_latency = 0
    else:
        refresh_latency = str(round(loop_duration_seconds - dolphie.refresh_interval, 2))

    use_performance_schema_status = "NO"
    if dolphie.use_performance_schema:
        use_performance_schema_status = "YES"

    if variables["read_only"] == "ON":
        if not dolphie.replica_status:
            variables["read_only"] = "YES ([red]SHOULD BE NO?[/red])"
        else:
            variables["read_only"] = "YES"
    elif variables["read_only"] == "OFF":
        variables["read_only"] = "NO"

    runtime = str(datetime.now() - dolphie.dolphie_start_time).split(".")[0]

    table_information.add_column()
    table_information.add_column(width=27)
    table_information.add_row("Name", "[grey93]%s" % dolphie.host, style=row_style)
    table_information.add_row("Version", "[grey93]%s %s" % (dolphie.host_distro, dolphie.full_version), style=row_style)
    table_information.add_row("Uptime", "[grey93]%s" % uptime, style=row_style)
    table_information.add_row(
        "Runtime", "[grey93]%s [grey78]latency: [grey93]%ss" % (runtime, refresh_latency), style=row_style
    )
    table_information.add_row("Read Only", "[grey93]%s" % variables["read_only"], style=row_style)
    table_information.add_row("Use PS", "[grey93]%s" % (use_performance_schema_status), style=row_style)
    table_information.add_row(
        "Threads",
        "[grey78]con[grey93] %s[steel_blue1]/[grey78]run[grey93] %s[steel_blue1]/[grey78]cac[grey93] %s"
        % (
            format_number(statuses["Threads_connected"]),
            format_number(statuses["Threads_running"]),
            format_number(statuses["Threads_cached"]),
        ),
        style=row_style,
    )
    table_information.add_row(
        "Tables",
        "[grey78]open[grey93] %s[steel_blue1]/[grey78]opened[grey93] %s"
        % (
            format_number(statuses["Open_tables"]),
            format_number(statuses["Opened_tables"]),
        ),
        style=row_style,
    )

    tables_to_add.append(table_information)

    ###########
    # InnoDB  #
    ###########
    table_innodb = Table(
        show_header=False,
        box=table_box,
        title="InnoDB",
        title_style=table_title_style,
        style=table_line_color,
    )

    table_innodb.add_column()
    table_innodb.add_column(width=9)

    # Calculate Checkpoint Efficiency
    innodb_log_file_size = variables["innodb_log_file_size"]

    # MariaDB Support
    if "innodb_log_files_in_group" in variables:
        innodb_log_files_in_group = variables["innodb_log_files_in_group"]
    else:
        innodb_log_files_in_group = 1

    # Save what percentage of log files InnoDB will start to aggressively flush
    # to disk due to checkpointing based on version
    if dolphie.full_version.startswith("8"):
        version_threshold = 0.875
    else:
        version_threshold = 0.81

    checkpoint_efficiency = "N/A"
    log_sequence_number_match = re.search(r"Log sequence number\s*(\d+)", innodb_status["status"])
    last_checkpoint_match = re.search(r"Last checkpoint at\s*(\d+)", innodb_status["status"])
    if log_sequence_number_match and last_checkpoint_match:
        checkpoint_age = int(log_sequence_number_match.group(1)) - int(last_checkpoint_match.group(1))

        max_checkpoint_age = innodb_log_file_size * innodb_log_files_in_group * version_threshold
        if checkpoint_age >= max_checkpoint_age:
            checkpoint_efficiency = "[bright_red]0.00%"
        elif max_checkpoint_age > checkpoint_age:
            checkpoint_efficiency = round(100 - (checkpoint_age / max_checkpoint_age * 100), 2)

            if checkpoint_efficiency > 40:
                checkpoint_efficiency = "[bright_green]%s%%" % checkpoint_efficiency
            elif checkpoint_efficiency > 20:
                checkpoint_efficiency = "[bright_yellow]%s%%" % checkpoint_efficiency
            else:
                checkpoint_efficiency = "[bright_red]%s%%" % checkpoint_efficiency

    # Get history list length
    output = re.search(r"History list length (\d+)", innodb_status["status"])
    if output:
        history_list_length = output.group(1)
    else:
        history_list_length = "N/A"

    # Calculate AIO reads
    total_pending_aio_reads = 0
    total_pending_aio_writes = 0
    output = re.search(
        r"Pending normal aio reads: (?:\d+\s)?\[(.*?)\] , aio writes: (?:\d+\s)?\[(.*?)\]",
        innodb_status["status"],
    )
    if output:
        match = output.group(1).split(",")

        for aio_read in match:
            total_pending_aio_reads += int(aio_read)

        match = output.group(2).split(",")
        for aio_write in match:
            total_pending_aio_writes += int(aio_write)
    else:
        total_pending_aio_reads = "N/A"
        total_pending_aio_writes = "N/A"

    # Calculate InnoDB memory read hit efficiency
    innodb_efficiency = "N/A"

    ib_pool_disk_reads = statuses["Innodb_buffer_pool_reads"]
    ib_pool_mem_reads = statuses["Innodb_buffer_pool_read_requests"]
    if ib_pool_disk_reads >= ib_pool_mem_reads:
        innodb_efficiency = "[bright_red]0.00%"
    elif ib_pool_mem_reads > ib_pool_disk_reads:
        innodb_efficiency = round(100 - (ib_pool_disk_reads / ib_pool_mem_reads * 100), 2)

        if innodb_efficiency > 90:
            innodb_efficiency = "[bright_green]%s%%" % innodb_efficiency
        elif innodb_efficiency > 80:
            innodb_efficiency = "[bright_yellow]%s%%" % innodb_efficiency
        else:
            innodb_efficiency = "[bright_red]%s%%" % innodb_efficiency

    # Calculate AHI Hit efficiency
    hash_searches = 0
    non_hash_searches = 0
    if variables["innodb_adaptive_hash_index"] == "ON":
        output = re.search(r"(\d+\.?\d+) hash searches\/s, (\d+\.?\d+) non-hash searches\/s", innodb_status["status"])
        if output:
            hash_searches = float(output.group(1))
            non_hash_searches = float(output.group(2))

            if non_hash_searches == 0 and hash_searches == 0:
                hash_search_efficiency = "Inactive"
            elif non_hash_searches >= hash_searches:
                hash_search_efficiency = "[bright_red]0.00%"
            elif hash_searches > non_hash_searches:
                hash_search_efficiency = round(100 - (non_hash_searches / hash_searches * 100), 2)

                if hash_search_efficiency > 70:
                    hash_search_efficiency = "[bright_green]%s%%" % hash_search_efficiency
                elif hash_search_efficiency > 50:
                    hash_search_efficiency = "[bright_yellow]%s%%" % hash_search_efficiency
                else:
                    hash_search_efficiency = "[bright_red]%s%%" % hash_search_efficiency
        else:
            hash_search_efficiency = "N/A"
    else:
        hash_search_efficiency = "OFF"

    # Get queries inside InnoDB
    output = re.search(r"(\d+) queries inside InnoDB, (\d+) queries in queue", innodb_status["status"])
    if output:
        queries_active = int(output.group(1))
        queries_queued = int(output.group(2))
    else:
        queries_active = "N/A"
        queries_queued = "N/A"

    # Calculate unpurged transactions
    output = re.search(r"Trx id counter (\d+)", innodb_status["status"])
    if output:
        trx_id_counter = int(output.group(1))
    else:
        trx_id_counter = None

    output = re.search(r"Purge done for trx's n:o < (\d+)", innodb_status["status"])
    if output:
        purge_done_for_trx = int(output.group(1))
    else:
        purge_done_for_trx = None

    if trx_id_counter is not None and purge_done_for_trx is not None:
        unpurged_trx = str(trx_id_counter - purge_done_for_trx)
    else:
        trx_id_counter = "N/A"

    # Add data to our table
    table_innodb.add_row("Read Hit", "[grey93]%s" % innodb_efficiency, style=row_style)
    table_innodb.add_row("Chkpt Age", "[grey93]%s" % checkpoint_efficiency, style=row_style)
    table_innodb.add_row("AHI Hit", "[grey93]%s" % (hash_search_efficiency), style=row_style)

    # Don't show thread concurrency information if it isn't set to on, instead show buffer pool stats
    if "innodb_thread_concurrency" in variables and variables["innodb_thread_concurrency"]:
        concurrency_ratio = round((queries_active / variables["innodb_thread_concurrency"]) * 100)

        if concurrency_ratio >= 80:
            queries_active_formatted = "[bright_red]%s" % format_number(queries_active)
        elif concurrency_ratio >= 60:
            queries_active_formatted = "[bright_yellow]%s" % format_number(queries_active)
        else:
            queries_active_formatted = "[grey93]%s" % format_number(queries_active)

        table_innodb.add_row(
            "Query Active",
            "[grey93]%s [steel_blue1]/ [grey93]%s" % (queries_active_formatted, variables["innodb_thread_concurrency"]),
            style=row_style,
        )
        table_innodb.add_row("Query Queued", "[grey93]%s" % format_number(queries_queued), style=row_style)
    else:
        table_innodb.add_row(
            "BP Size",
            "[grey93]%s" % (format_bytes(float(variables["innodb_buffer_pool_size"]))),
            style=row_style,
        )
        table_innodb.add_row(
            "BP Dirty",
            "[grey93]%s" % (format_bytes(float(statuses["Innodb_buffer_pool_bytes_dirty"]))),
            style=row_style,
        )

    if str(total_pending_aio_reads) == "N/A":
        table_innodb.add_row(
            "Pending AIO",
            "[gray93]N/A",
            style=row_style,
        )
    else:
        table_innodb.add_row(
            "Pending AIO",
            "[gray78]W [grey93]%s [gray78]R [grey93]%s" % (str(total_pending_aio_writes), str(total_pending_aio_reads)),
            style=row_style,
        )

    table_innodb.add_row(
        "History List",
        "[grey93]%s" % format_number(history_list_length),
        style=row_style,
    )
    table_innodb.add_row("Unpurged TRX", "[grey93]%s" % format_number(unpurged_trx), style=row_style)

    tables_to_add.append(table_innodb)

    ##############
    # Binary Log #
    ##############
    table_primary = Table()

    if primary_status:
        table_primary = Table(
            show_header=False,
            box=table_box,
            title="Binary Log",
            title_style=table_title_style,
            style=table_line_color,
        )

        if dolphie.previous_binlog_position == 0:
            diff_binlog_position = 0
        elif dolphie.previous_binlog_position > primary_status["Position"]:
            diff_binlog_position = "Binlog Rotated"
        else:
            diff_binlog_position = format_bytes(primary_status["Position"] - dolphie.previous_binlog_position)

        binlog_cache = 100
        binlog_cache_disk = statuses["Binlog_cache_disk_use"]
        binlog_cache_mem = statuses["Binlog_cache_use"]
        if binlog_cache_disk and binlog_cache_mem:
            if binlog_cache_disk >= binlog_cache_mem:
                innodb_efficiency = "[bright_red]0.00%"
            elif binlog_cache_mem > binlog_cache_disk:
                binlog_cache = round(100 - (binlog_cache_disk / binlog_cache_mem), 2)

        table_primary.add_column()
        table_primary.add_column(max_width=40)
        table_primary.add_row("File name", "[grey93]%s" % str(primary_status["File"]), style=row_style)
        table_primary.add_row(
            "Position",
            "[grey93]%s" % (str(primary_status["Position"])),
            style=row_style,
        )
        table_primary.add_row(
            "Size",
            "[grey93]%s" % format_bytes(primary_status["Position"]),
            style=row_style,
        )
        table_primary.add_row("Diff", "[grey93]%s" % diff_binlog_position, style=row_style)
        table_primary.add_row("Cache Hit", "[grey93]%s%%" % str(binlog_cache), style=row_style)
        # MariaDB Support
        if "gtid_mode" in variables:
            table_primary.add_row("GTID", "[grey93]%s" % str(variables["gtid_mode"]), style=row_style)
        else:
            table_primary.add_row()
        table_primary.add_row()
        table_primary.add_row()

        tables_to_add.append(table_primary)

    ###############
    # Replication #
    ###############
    if dolphie.replica_status and dolphie.layout["replicas"].visible is False:
        tables_to_add.append(replica_panel.create_table(dolphie, dolphie.replica_status, dashboard_table=True))

    ###############
    # Statisitics #
    ###############
    table_stats = Table(
        show_header=False,
        box=table_box,
        title="Statisitics/s",
        title_style=table_title_style,
        style=table_line_color,
    )

    table_stats.add_column()
    table_stats.add_column(min_width=7)

    if loop_duration_seconds == 0:
        queries_per_second = 0
        connects_per_second = 0
        selects_per_second = 0
        inserts_per_second = 0
        updates_per_second = 0
        deletes_per_second = 0
        replaces_per_second = 0
        rollbacks_per_second = 0
    else:
        queries_per_second = round((statuses["Queries"] - saved_status["Queries"]) / loop_duration_seconds)

        connects_per_second = round((statuses["Connections"] - saved_status["Connections"]) / loop_duration_seconds)
        selects_per_second = round((statuses["Com_select"] - saved_status["Com_select"]) / loop_duration_seconds)
        inserts_per_second = round((statuses["Com_insert"] - saved_status["Com_insert"]) / loop_duration_seconds)
        updates_per_second = round((statuses["Com_update"] - saved_status["Com_update"]) / loop_duration_seconds)
        deletes_per_second = round((statuses["Com_delete"] - saved_status["Com_delete"]) / loop_duration_seconds)
        replaces_per_second = round((statuses["Com_replace"] - saved_status["Com_replace"]) / loop_duration_seconds)
        rollbacks_per_second = round((statuses["Com_rollback"] - saved_status["Com_rollback"]) / loop_duration_seconds)

    table_stats.add_row(
        "Queries",
        "[grey93]%s" % format_number(queries_per_second),
        style=row_style,
    )
    table_stats.add_row("SELECT", "[grey93]%s" % format_number(selects_per_second), style=row_style)
    table_stats.add_row("INSERT", "[grey93]%s" % format_number(inserts_per_second), style=row_style)
    table_stats.add_row("UPDATE", "[grey93]%s" % format_number(updates_per_second), style=row_style)
    table_stats.add_row("DELETE", "[grey93]%s" % format_number(deletes_per_second), style=row_style)
    table_stats.add_row(
        "REPLACE",
        "[grey93]%s" % format_number(replaces_per_second),
        style=row_style,
    )
    table_stats.add_row(
        "ROLLBACK",
        "[grey93]%s" % format_number(rollbacks_per_second),
        style=row_style,
    )
    table_stats.add_row(
        "CONNECT",
        "[grey93]%s" % format_number(connects_per_second),
        style=row_style,
    )

    tables_to_add.append(table_stats)

    dashboard_grid.add_row(*tables_to_add)

    dashboard_panel = Panel(
        Align.center(dashboard_grid),
        box=box.SIMPLE,
        border_style="steel_blue1",
    )

    return dashboard_panel



import pymysql


class Database:
    def __init__(self, host, user, password, socket, port, ssl):
        self.host = host
        self.user = user
        self.password = password
        self.socket = socket
        self.port = port
        self.ssl = ssl

        try:
            self.connection = pymysql.connect(
                host=host,
                user=user,
                passwd=password,
                unix_socket=socket,
                port=port,
                use_unicode=False,
                ssl=ssl,
                autocommit=True,
            )
            self.cursor = self.connection.cursor(pymysql.cursors.DictCursor)
        except pymysql.Error as e:
            raise Exception("Failed to connect to database host %s - Reason: %s" % (self.host, e.args[1]))
        except FileNotFoundError:  # Catch SSL file path errors
            raise Exception("SSL certificate file path isn't valid!")

    def fetchone(self, query, field, values=None):
        if values is None:
            values = []
        self.cursor.execute(query, tuple(values))
        data = self.cursor.fetchone()
        if isinstance(data[field], (bytes, bytearray)):
            return data[field].decode()
        else:
            return data[field]

    def close(self):
        if self.connection:
            self.connection.close()



import atexit
import sys
import termios
from select import select


class KBHit:
    def __init__(self):
        """
        Creates a KBHit object that you can call to do various keyboard things.
        """

        # Save the terminal settings
        self.fd = sys.stdin.fileno()
        self.new_term = termios.tcgetattr(self.fd)
        self.old_term = termios.tcgetattr(self.fd)

        # New terminal setting unbuffered
        self.new_term[3] = self.new_term[3] & ~termios.ICANON & ~termios.ECHO
        termios.tcsetattr(self.fd, termios.TCSANOW, self.new_term)

        # Support normal-terminal reset at exit
        atexit.register(self.set_normal_term)

    def set_new_term(self):
        """
        Sets terminal up to catch keys
        """

        self.new_term[3] = self.new_term[3] & ~termios.ICANON & ~termios.ECHO
        termios.tcsetattr(self.fd, termios.TCSANOW, self.new_term)

    def set_normal_term(self):
        """
        Resets to normal terminal.
        """

        termios.tcsetattr(self.fd, termios.TCSANOW, self.old_term)

    def getch(self):
        """
        Returns a keyboard character after kbhit() has been called.
        Should not be called in the same program as getarrow().
        """

        return sys.stdin.read(1)

    def getarrow(self):
        """
        Returns an arrow-key code after kbhit() has been called. Codes are
        0 : up
        1 : right
        2 : down
        3 : left
        Should not be called in the same program as getch().
        """

        c = sys.stdin.read(3)[2]
        vals = [65, 67, 66, 68]

        return vals.index(ord(c.decode("utf-8")))

    def key_press(self):
        """
        Returns True if keyboard character was hit, False otherwise.
        """

        dr, dw, de = select([sys.stdin], [], [], 0)
        return dr != []

    def key_press_blocking(self):
        self.set_new_term()

        try:
            return sys.stdin.read(1)
        except IOError:
            return 0
